# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Data source URL"""

import logging
import urlparse
try:
    from urlparse import parse_qsl
except ImportError:
    from cgi import parse_qsl

import cf.db

def _initialize():
    # Customize urlparse
    global _intialized
    if _initialized:
        return
    [urlparse.uses_netloc.append(drivername)
     for drivername in cf.db.DIALECTS
     if drivername not in urlparse.uses_netloc]

    [urlparse.uses_query.append(drivername)
     for drivername in cf.db.DIALECTS
     if drivername not in urlparse.uses_query]

    _intialized = True

_initialized = False


class URL(object):

    def __init__(self, drivername, user=None, password=None,
                 host=None, port=None, database=None, **kwds):
        self.drivername = drivername
        self.user = user
        self.password = password
        self.host = host
        self.port = self._clean_port(port)
        self.database = database
        self.query = kwds
        _initialize()

    def _get_username(self):
        return self.user

    def _set_username(self, username):
        self.user = username

    username = property(fget=_get_username, fset=_set_username)

    def __str__(self):
        netloc = ''
        if self.user:
            netloc += self.user
        if self.password:
            netloc += ':%s' % self.password
        if self.user:
            netloc += '@'
        if self.host:
            netloc += self.host
        if self.port:
            netloc += ':%s' % self.port
        if self.database and self.database.startswith('/'):
            netloc += '/'
        if self.query:
            query = '%s' % '&'.join(['%s=%s' % (k, v)
                                     for k, v in self.query.iteritems()])
        else:
            query = None
        return urlparse.urlunsplit((self.drivername, netloc,
                                    self.database or '', query, None))

    @classmethod
    def _clean_port(cls, port):
        """Return port as integer or None."""
        if port is None:
            return port
        if not isinstance(port, int):
            try:
                port = int(port)
            except (TypeError, ValueError), err:
                port = None
                logging.exception('Failed to convert port to int %r: %s',
                                  port, err)
        return port

    def get_dict(self):
        """Return instance values as dictionary."""
        data = {}
        if self.user:
            data['username'] = self.user
        if self.password:
            data['password'] = self.password
        if self.host:
            data['host'] = self.host
        if self.port:
            data['port'] = self.port
        if self.database:
            data['database'] = self.database
        data.update(self.query)
        return data


def make_url(raw):
    """Create URL from string."""
    _initialize()
    parsed = urlparse.urlsplit(raw)
    if (parsed.path and parsed.path.startswith('/')
        and not parsed.path.startswith('//')):
        database = parsed.path[1:]
    else:
        database = parsed.path or None
    if database and database.startswith('//'):
        database = database[1:]
    if parsed.query:
        kwds = dict(parse_qsl(parsed.query))
    else:
        kwds = {}
    # The following try-except block cleans up bad data that may life
    # in the datasource configuration file (see issue70).
    try:
        port = parsed.port
    except ValueError:
        port = None
    return URL(parsed.scheme,
               parsed.username or None, parsed.password or None,
               parsed.hostname or None, port or None,
               database, **kwds)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Oracle backend"""

from gettext import gettext as _

from cf.db import objects
from cf.db.backends import Generic, GUIOption


class Oracle(Generic):

    drivername = 'oracle'

    @classmethod
    def dbapi(cls):
        import cx_Oracle
        return cx_Oracle

    @classmethod
    def get_native_shell_command(cls, url):
        args = []
        logon = ''
        if url.user:
            logon += url.user
        if url.password:
            logon += '/%s' % url.password
        if url.host:
            logon += '@%s' % url.host
        if 'mode' in url.query and url.query['mode']:
            logon += ' AS %s' % url.query['mode']
        args.extend([logon])
        return 'sqlplus', args

    @classmethod
    def get_connect_params(cls, url):
        # derived from sqlalchemy.databases.oracle
        if url.database:
            # if we have a database, then we have a remote host
            port = url.port
            if port:
                port = int(port)
            else:
                port = 1521
            dsn = cls.dbapi().makedsn(url.host, port, url.database)
        else:
            # we have a local tnsname
            dsn = url.host

        opts = dict(user=url.username, password=url.password, dsn=dsn)
        if 'mode' in url.query:
            opts['mode'] = url.query['mode']
            if isinstance(opts['mode'], basestring):
                mode = opts['mode'].upper()
                if mode == 'SYSDBA':
                    opts['mode'] = cls.dbapi().SYSDBA
                elif mode == 'SYSOPER':
                    opts['mode'] = cls.dbapi().SYSOPER
                else:
                    opts['mode'] = int(opts['mode'])
        return tuple(), opts

    @classmethod
    def get_options(cls):
        return (
            GUIOption('host', _(u'TNS Name')),
            GUIOption('username', _(u'Username')),
            GUIOption('password', _(u'Password'),
                      widget=GUIOption.WIDGET_PASSWORD),
            GUIOption('mode', _(u'Mode'),
                      widget=GUIOption.WIDGET_COMBO,
                      choices=(
                          (None, None),
                          ('SYSDBA', 'SYSDBA'),
                          ('SYSOPER', 'SYSOPER'),
                      )),
            )

    @classmethod
    def get_error_position(cls, query, error):
        if not isinstance(error, cls.dbapi().DatabaseError):
            return None
        ora_error = error.args[0]
        if not isinstance(ora_error, cls.dbapi()._Error):
            return None
        offset = getattr(ora_error, 'offset', None)
        if offset is None:
            return None
        loffset = new_offset = 0
        for idx, line in enumerate(query.statement.splitlines(True)):
            new_offset += len(line)
            if new_offset > offset:
                line_offset = offset-loffset
                return idx+1, line_offset+1
            loffset = new_offset
        return None

    def get_server_info(self, connection):
        return 'Oracle %s' % connection.connection.version

    def prepare_statement(self, sql):
        # See issue50: cx_Oracle requires str or None for cursor.execute().
        sql = str(sql)
        # Another issue: Somehow Oracle dislikes trailing semicolons.
        # Maybe this approach is a little bit crude, but just remove it...
        # See this thread for example:
        # http://www.mail-archive.com/django-users@googlegroups.com/msg11479.html
        sql = sql.strip()
        if sql.endswith(';'):
            sql = sql[:-1]
        return str(sql)

    def get_explain_statement(self, statement):
        return ['explain plan for %s' % statement, 'select * from plan_table']

    def _query(self, connection, sql):
        return connection.execute_raw(sql)

    def initialize(self, meta, connection):
        schemata = objects.Schemata(meta)
        users = objects.Users(meta)
        meta.set_object(schemata)
        meta.set_object(users)
        tbl_cache = {}  # speed up object creation
        for item in self._query(connection, INITIAL_SQL):
            if item['TYPE'] == 'schema':
                schema = objects.Schema(meta, oid=item['ID'],
                                        name=item['NAME'], parent=schemata)
                meta.set_object(schema)
                tables = objects.Tables(meta, parent=schema)
                meta.set_object(tables)
                schema.tables = tables
                views = objects.Views(meta, parent=schema)
                meta.set_object(views)
                schema.views = views
                schema.sequences = objects.Sequences(meta, parent=schema)
                meta.set_object(schema.sequences)

            if item['TYPE'] in ('table', 'view'):
                schema = meta.find_exact(cls=objects.Schema,
                                         oid=item['PARENT'])
                if item['TYPE'] == 'table':
                    cls = objects.Table
                    parent = meta.find_exact(cls=objects.Tables, parent=schema)
                else:
                    cls = objects.View
                    parent = meta.find_exact(cls=objects.Views, parent=schema)
                obj = cls(meta, parent=parent, oid=item['ID'],
                          name=item['NAME'], comment=item['DESCRIPTION'])
                meta.set_object(obj)
                tbl_cache[item['ID']] = obj
                if item['TYPE'] == 'table':
                    obj.indexes = objects.Indexes(meta, parent=obj)
                    meta.set_object(obj.indexes)
                obj.triggers = objects.Triggers(meta, parent=obj)
                meta.set_object(obj.triggers)

            if item['TYPE'] == 'column':
                parent = tbl_cache.get(item['PARENT'])
                if parent is None:
                    continue
                col = objects.Column(meta, parent=parent.columns,
                                     oid=item['ID'],
                                     name=item['NAME'],
                                     comment=item['DESCRIPTION'])
                meta.set_object(col)


    def refresh(self, obj, meta, connection):
        if obj.typeid == 'users':
            self._refresh_users(obj, meta, connection)
        elif obj.typeid == 'constraints':
            self._refresh_constraints(obj, meta, connection)
        elif obj.typeid == 'indexes':
            self._refresh_indexes(obj, meta, connection)
        elif obj.typeid == 'triggers':
            self._refresh_triggers(obj, meta, connection)
        elif obj.typeid == 'sequences':
            self._refresh_sequences(obj, meta, connection)

    def _refresh_users(self, coll, meta, connection):
        sql = "select username from sys.all_users"
        for item in self._query(connection, sql):
            u = meta.find_exact(parent=coll, oid=item['USERNAME'].lower())
            if u is None:
                u = objects.User(meta, parent=coll,
                                 oid=item['USERNAME'].lower(),
                                 name=item['USERNAME'])
                meta.set_object(u)
                u.props.has_children = False

    def _refresh_constraints(self, coll, meta, connection):
        sql = ("select lower(owner||'.'||constraint_name) as id,"
               " constraint_name as name"
               " from sys.all_constraints"
               " where lower(owner||'.'||table_name) = '%s'"
               % coll.parent.oid)
        for item in self._query(connection, sql):
            con = meta.find_exact(parent=coll, oid=item['ID'])
            if con is None:
                con = objects.Constraint(meta, parent=coll,
                                         oid=item['ID'])
                meta.set_object(con)
            con.name = item['NAME']
            con.pros.has_children = False

    def _refresh_indexes(self, coll, meta, connection):
        sql = ("select lower(owner||'.'||index_name) as id,"
               " index_name as name from sys.all_indexes"
               " where lower(owner||'.'||table_name) = '%s'"
               % coll.parent.oid)
        for item in self._query(connection, sql):
            con = meta.find_exact(parent=coll, oid=item['ID'])
            if con is None:
                con = objects.Index(meta, parent=coll,
                                    oid=item['ID'])
                meta.set_object(con)
            con.name = item['NAME']
            con.props.has_children = False

    def _refresh_triggers(self, coll, meta, connection):
        sql = ("select lower(owner||'.'||trigger_name) as id,"
               " trigger_name as name"
               " from sys.all_triggers"
               " where lower(owner||'.'||table_name) = '%s'"
               % coll.parent.oid)
        for item in self._query(connection, sql):
            con = meta.find_exact(parent=coll, oid=item['ID'])
            if con is None:
                con = objects.Trigger(meta, parent=coll,
                                      oid=item['ID'])
                meta.set_object(con)
            con.name = item['NAME']
            con.props.has_children = False

    def _refresh_sequences(self, coll, meta, connection):
        sql = ("select lower(sequence_owner||'.'||sequence_name) as id,"
               " sequence_name as name"
               " from sys.all_sequences"
               " where lower(sequence_owner) = '%s'"
               % coll.parent.oid)
        for item in self._query(connection, sql):
            seq = meta.find_exact(parent=coll, oid=item['ID'])
            if seq is None:
                seq = objects.Sequence(meta, parent=coll, oid=item['ID'])
                meta.set_object(seq)
            seq.name = item['NAME']
            seq.props.has_children = False


DRIVER = Oracle


INITIAL_SQL = """select * from (
select lower(username) as id, null as parent,
username as name, null as description, 'schema' as type, 1 as pos
from sys.all_users
where exists (select 'x' from sys.all_objects where owner=username) or username = user

union

select
lower(t.owner||'.'||t.table_name) as id, lower(t.owner) as parent,
t.table_name as name, c.comments as description, 'table' as type, 2 as pos
from sys.all_tables t
left join all_tab_comments c on c.owner = t.owner
and c.table_name = t.table_name
and c.table_type = 'TABLE'

union

select
lower(t.owner||'.'||t.view_name) as id, lower(t.owner) as parent,
t.view_name as name, c.comments as description, 'table' as type, 3 as pos
from sys.all_views t
left join all_tab_comments c on c.owner = t.owner
and c.table_name = t.view_name
and c.table_type = 'VIEW'

union

select
lower(t.owner||'.'||t.table_name||'.'||t.column_name) as id,
lower(t.owner||'.'||t.table_name) as parent,
t.column_name as name, c.comments as description, 'column' as type, 4 as pos
from sys.all_tab_columns t
left join sys.all_col_comments c on c.owner = t.owner
and c.table_name = t.table_name and c.column_name = t.column_name
) x order by pos, name"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Database meta information."""

from gettext import gettext as _

import logging
import thread

import gobject
import gtk




class DatabaseMeta(object):

    def __init__(self, datasource):
        self.datasource = datasource
        self.app = datasource.manager.app
        self.conn = self.datasource.internal_connection
        self._items = set()
        if self.conn.threadsafety >= 2:
            thread.start_new_thread(self.initialize, (True,))
        else:
            self.initialize()

    @property
    def backend(self):
        return self.datasource.backend

    def get_server_info(self):
        return self.backend.get_server_info(self.conn)

    def initialize(self, threaded=False):
        """Called after instance creation to prepare intial data."""
        if threaded:
            gtk.gdk.threads_enter()
        self.app.set_status_message(_(u'Loading database structure...'), 100)
        if threaded:
            gtk.gdk.threads_leave()
        try:
            self.backend.initialize(self, self.conn)
        except:
            msg = 'DatabaseMeta.initialize failed (driver: %s):'
            msg = msg % self.backend.drivername
            logging.exception(msg)
        if threaded:
            gtk.gdk.threads_enter()
        self.app.pop_status_message(100)
        if threaded:
            gtk.gdk.threads_leave()

    def set_object(self, obj):
        """Adds or replaces an object."""
        if obj in self._items:
            self._items.remove(obj)
        self._items.add(obj)

    def get_children(self, parent=None):
        """Get child objects for parent."""
        if parent is not None and parent.props.refresh_required:
            self.backend.refresh(parent, self, self.conn)
            parent.props.refresh_required = False
        return self.find(parent=parent)

    def find(self, **kwds):
        """Find an object using searchterm."""
        res = list(self._items)
        if 'cls' in kwds:
            objcls = kwds.pop('cls')
            res = filter(lambda x: isinstance(x, objcls), res)
        for key, value in kwds.iteritems():
            res = filter(lambda x: x.property_matches(key, value), res)
        return res

    def find_exact(self, **kwds):
        """Like :meth:`find`, but returns exactly one match or ``None``."""
        res = self.find(**kwds)
        if len(res) > 1:
            raise RuntimeError('More than one match found (%d)' % len(res))
        elif len(res) == 1:
            return res[0]
        return None



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Instance selector class"""

import os.path

import gtk

import cf


class InstanceSelector(object):

    def __init__(self, client):
        self.client = client
        self.builder = gtk.Builder()
        self.builder.set_translation_domain('crunchyfrog')
        fname = os.path.join(cf.DATA_DIR, 'glade/instanceselector.glade')
        self.builder.add_from_file(fname)
        self.builder.connect_signals(self)
        self.dlg = self.builder.get_object('instanceselector')
        self._setup_widget()
        self._setup_connections()

    def _setup_widget(self):
        model = gtk.ListStore(int, str)
        self.list = self.builder.get_object("list_instances")
        self.list.set_model(model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), text=1)
        self.list.append_column(col)
        for id, title in self.client.get_instances():
            model.append([id, title])

    def _setup_connections(self):
        sel = self.list.get_selection()
        sel.connect("changed", self.on_isel_changed)
        self.builder.get_object("btn_newinstance").connect(
            "toggled", self.on_newi_toggled)

    # ---------
    # Callbacks
    # ---------

    def on_newi_toggled(self, btn):
        sel = self.list.get_selection()
        model = self.list.get_model()
        if btn.get_active():
            sel.unselect_all()
        else:
            sel.select_iter(model.get_iter_first())
        self.list.set_sensitive(not btn.get_active())

    def on_isel_changed(self, selection):
        model, iter = selection.get_selected()
        if not iter:
            self.builder.get_object("btn_newinstance").set_active(True)
        else:
            self.builder.get_object("btn_activeinstance").set_active(True)

    # --------------
    # Public methods
    # --------------

    def run(self):
        """Run the dialog."""
        return self.dlg.run()

    def destroy(self):
        """Destroy the dialog."""
        return self.dlg.destroy()

    def get_instance_id(self):
        if self.builder.get_object("btn_newinstance").get_active():
            return None
        else:
            sel = self.list.get_selection()
            model, iter = sel.get_selected()
            if iter:
                return model.get_value(iter, 0)
            return None



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line interface."""

import logging
import optparse
import os
import sys

import cf
from cf import release


def get_parser():
    """Creates and returns the command line parser."""
    usage = "usage: %prog [options] FILE1, FILE2, ..."
    parser = optparse.OptionParser(usage, prog=release.appname,
                                   version=release.version)
    parser.add_option("-c", "--config",
                      dest="config", default=cf.USER_CONF,
                      help="configuration file")
    group = parser.add_option_group('Logging options')
    group.add_option('-q', '--quiet', action='store_const', const=0,
                     dest='verbose', help='Print errors only.')
    group.add_option('-v', '--verbose', action='store_const', const=2,
                     dest='verbose', default=1,
                     help='Print info level logs (default).')
    group.add_option('--noisy', action='store_const', const=3,
                     dest='verbose', help='Print all logs.')
    return parser


def run():
    """Runs the application."""
    logging.basicConfig(format=('%(created)f %(levelname)s %(filename)s:'
                                '%(lineno)s %(message)s '))

    parser = get_parser()
    opts, args = parser.parse_args()

    if opts.verbose >= 3:
        logging.getLogger().setLevel(logging.DEBUG)
    elif opts.verbose >= 2:
        logging.getLogger().setLevel(logging.INFO)

    opts.first_run = not os.path.isfile(opts.config)

    try:
        from cf import ipc
        ipc_client = ipc.get_client()
    except AttributeError, err:
        logging.warning('No IPC available: %s', err)
        ipc_client = None

    # Let's import gtk and related.
    # Note: We import them here to allow command line usage
    # (e.g. crunchyfrog --version) without gkt requirements.
    try:
        import glib
        glib.set_application_name('CrunchyFrog')
    except ImportError:
        logging.debug('glib module not found.')

    import pygtk
    if not sys.platform == 'win32':
        pygtk.require('2.0')
    import gtk

    try:
        import gnome
        have_gnome = True
    except ImportError:
        have_gnome = False
    logging.info('GNOME enabled: %s', have_gnome and 'yes' or 'no')

    from cf.app import CFApplication

    # Check for running applications or create one
    if ipc_client is None or not ipc.is_alive(ipc_client):
        logging.info('Creating new application')
        if have_gnome:
            props = {}
            setup_py = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                    '../setup.py'))
            if os.path.isfile(setup_py):
                props['app-datadir'] = cf.DATA_DIR
            props['human-readable-name'] = release.name
            gnome.init(release.name.lower(), release.version, properties=props)
        gtk.gdk.threads_init()
        gtk.gdk.threads_enter()
        app = CFApplication(opts)
        app.init()
        instance = app.new_instance(args)
        logging.info('Entering GTK main loop')
        gtk.main()
        logging.info('Leaving GTK main loop')
        gtk.gdk.threads_leave()

    # We have a running application, re-use it...
    else:
        logging.info('Running application found')
        if args:
            # TODO(andi): Move instance selector to app.
            #             - It causes broken pipes on socket.
            #             - If it's in app it can have a mainwin as parent.
            from cf.instance import InstanceSelector
            sel = InstanceSelector(ipc_client)
            if sel.run() == 1:
                instance_id = sel.get_instance_id()
                if not instance_id:
                    instance_id = ipc_client.new_instance(args)
                for fname in args:
                    ipc_client.open_uri(instance_id, fname)
            sel.destroy()
        else:
            ipc_client.new_instance()

if __name__ == '__main__':
    run()


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Autocompletion support for SQL editor.

The functions in this module provide autocompletion support for SQL editor
widgets.

The feature is enabled by calling the :meth:`setup` function. It then
automatically connects to instance and editor creation and to relevant
config changes.
"""

import itertools
from gettext import gettext as _

import gobject
import gtk
import pango

import sqlparse.keywords

from cf import db


SQL_KEYWORDS = set(tuple(sqlparse.keywords.KEYWORDS))
SQL_KEYWORDS.update(tuple(sqlparse.keywords.KEYWORDS_COMMON))


def setup(app):
    """Setup autocompletion feature.

    Connect to 'instance-created' and 'changed' (config) signal.

    :param app: Application instance.
    """
    app.cb.connect('instance-created', instance_created, app)
    app.config.connect('changed', on_config_changed)


def instance_created(callbacks, window, app):
    """Adds UI definition and actions to the new instance."""
    group = [x for x in window.ui.get_action_groups()
             if x.get_name() == 'editor'][0]
    group.add_actions([('query-autocomplete', None,
                        _(u'_Auto-Complete'), '<control>space',
                        None, run_action)], window)
    window.ui.add_ui_from_string(UI)
    window.connect('editor-created', on_editor_created)


def run_action(action, window):
    """Run the autocompletion action."""
    editor = window.get_active_editor()
    if editor is None:
        return
    if editor.textview.get_data('cf::ac_window') is not None:
        return
    editor_autocomplete(editor)


def editor_autocomplete(editor, popup=None, matches=None):
    """Implementation of the autocompletion feature.

    :param editor: SQLEditor instance.
    :param popup: The popup window. If `None` (default) a new window is
      created.
    :param matches: If given, the completions are set from these matches.
      Otherwise the possible completions are build within this function.
      The default is `None`.
    """
    if matches is None:
        matches = get_matches(editor)
    if matches is None:
        return
    scores = matches.keys()
    if not scores:
        if popup is not None:
            destroy_popup(editor.textview, popup)
        return
    buffer_ = editor.textview.buffer
    start, end = get_fragment_bounds(buffer_)
    iter_rect = editor.textview.get_iter_location(end)
    x, y = editor.textview.window.get_origin()
    y2, height = editor.textview.get_line_yrange(end)
    left_window = editor.textview.get_window(gtk.TEXT_WINDOW_LEFT)
    if popup is None:
        popup = editor.textview.get_data('cf::ac_window')
        if popup is None:
            popup = create_popup_window(editor)
    model = popup.get_data('model')
    treeview = popup.get_data('treeview')
    popup.move(x+iter_rect.x+left_window.get_size()[0], y+iter_rect.y+height)
    populate_model(model, matches)
    iter_ = model.get_iter_first()
    selection = treeview.get_selection()
    selection.select_iter(iter_)


def _is_keyword(value):
    """Return True if value is a SQL keyword."""
    return value.upper() in SQL_KEYWORDS


def editor_autocomplete_advanced(editor):
    """Implementation of advanced autocompletion (tab-completion).

    If there's exactly one match, the match is applied automatically.
    Otherwise the usual :meth:`editor_autocompletion` implementation is
    called.

    :param editor: SQLEditor instance.
    """
    matches = get_matches(editor)
    if matches is None:
        return False
    completions = set()
    for value in matches.itervalues():
        for compl in value:
            completions.add(compl[1][0])
    if len(completions) == 0:
        return True
    elif len(completions) != 1:
        editor_autocomplete(editor, matches=matches)
        return True
    value = completions.pop()
    _apply_selection(editor, value, add_blank=_is_keyword(value))
    return True


# ----
# Completion Popup
# ----

def create_popup_window(editor):
    """Create a new popup window for an editor.

    The created popup window is set as widget data to *editor*.

    :param editor: SQLEditor instance.
    """
    w = gtk.Window(gtk.WINDOW_POPUP)
    w.set_name('gtk-tooltips')
    w.set_border_width(4)
    w.set_decorated(False)
    w.set_app_paintable(True)
    w.connect('expose-event', popup_window_expose)
    cpl_list = gtk.TreeView()
    cpl_list.set_headers_visible(False)
    model = gtk.ListStore(str, str, str, str)
    model.set_data('cf::bgcolor1', w.get_style().bg[gtk.STATE_NORMAL])
    model.set_data('cf::bgcolor2', w.get_style().bg[gtk.STATE_PRELIGHT])
    cpl_list.set_model(model)
    cpl_list.set_style(w.get_style())
    col = gtk.TreeViewColumn(None, gtk.CellRendererText(), markup=0,
                             background=2)
    col.set_expand(True)
    cpl_list.append_column(col)
    col = gtk.TreeViewColumn(None, gtk.CellRendererText(), markup=1,
                             background=2)
    col.set_expand(False)
    cpl_list.append_column(col)
    sw = gtk.ScrolledWindow()
    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
    sw.add(cpl_list)
    w.add(sw)
    w.resize(350, 160)
    w.show_all()
    w.set_data('model', model)
    w.set_data('treeview', cpl_list)
    textview = editor.textview
    textview.set_data('cf::ac_window', w)
    sigs = []
    sigs.append(textview.connect('key-press-event',
                                 editor_key_pressed, editor, w))
    sigs.append(textview.connect('delete-event',
                                 lambda x, e: destroy_popup(textview, w)))
    sigs.append(textview.connect('button-press-event',
                                 lambda x, e: destroy_popup(textview, w)))
    sigs.append(textview.connect('scroll-event',
                                 lambda x, e: destroy_popup(textview, w)))
    sigs.append(textview.connect('focus-out-event',
                                 lambda x, e: destroy_popup(textview, w)))
    textview.set_data('cf::autocomplete_sigs', sigs)
    return w


def popup_window_expose(window, event):
    """Draw a flat box around the popup window on expose event.

    This makes it more look like a 'real' tooltip.
    """
    w, h = window.get_size()
    window.style.paint_flat_box(window.window, gtk.STATE_NORMAL,
                                gtk.SHADOW_OUT, None, window,
                                'tooltip', 0, 0, w, h)


def destroy_popup(textview, window):
    """Destroy and unregister the popup window."""
    sigs = textview.get_data('cf::autocomplete_sigs')
    if sigs is None:
        return
    for sig in sigs:
        textview.disconnect(sig)
    textview.set_data('cf::autocomplete_sigs', None)
    textview.set_data('cf::ac_window', None)
    window.destroy()


def populate_model(model, matches):
    """Populate a ListStore with matches."""
    model.clear()
    color1 = model.get_data('cf::bgcolor1')
    color2 = model.get_data('cf::bgcolor2')
    colors = itertools.cycle([color1, color2])
    scores = list(matches)
    scores.sort()
    for score in scores:
        xmatches = matches[score]
        xmatches.sort(key=lambda x: x[1][0])
        for match in xmatches:
            hint = '<i>%s</i>' % gobject.markup_escape_text(match[1][1])
            sol = ''
            for idx, char in enumerate(match[1][0]):
                if idx in match[0]:
                    tmp = '<b>%s</b>'
                else:
                    tmp = '%s'
                sol += tmp % gobject.markup_escape_text(char)
            model.append([sol, hint, colors.next(), match[1][0]])


def select_offset(popup, offset):
    """Select another item in the completion list.

    :param popup: The popup window.
    :param offset: Item index relative to the selected item.
    """
    treeview = popup.get_data('treeview')
    selection = treeview.get_selection()
    model, iter_ = selection.get_selected()
    if iter_ is None:
        iter_ = model.get_iter_first()
    row = model.get_path(iter_)[0]
    row += offset
    if row < 0:
        row = 0
    try:
        iter_ = model.get_iter(row)
    except ValueError:
        # end of completion list reached.
        # maybe this could be catched earlier be looking at the actual
        # number of rows.
        iter_ = model.get_iter_first()
    selection.select_iter(iter_)
    treeview.scroll_to_cell(row)


def apply_selection(popup, editor):
    """Applies the selected item."""
    treeview = popup.get_data('treeview')
    selection = treeview.get_selection()
    model, iter_ = selection.get_selected()
    value = model.get_value(iter_, 3)
    _apply_selection(editor, value)
    destroy_popup(editor.textview, popup)


# ----
# Helper functions to find and build the possible completions
# ----

def build_completions(editor, fragment):
    """Build the common completions.

    The returned list contains the names of database objects if the editor
    has a connection and meta information. Additionally the list contains
    SQL keywords.

    The returned list is a list of 2-tuples (completion, description) where
    description describes the object (e.g. 'Keyword', 'Table', 'Column'...).

    :param editor: SQLEditor instance.
    :param fragment: The fragment that should be completed.
    """
    ret = []
    # objects
    if editor.connection and editor.connection.meta:
        meta = editor.connection.meta
        [ret.append((obj.get_full_name(), obj.typestr))
         for obj in meta.find(name__ilike=fragment,
                              cls=(db.objects.Table,
                                   db.objects.View,
                                   db.objects.Sequence,
                                   db.objects.Schema))]
    # keywords
    kwds = tuple(sqlparse.keywords.KEYWORDS_COMMON)
    kwds += tuple(sqlparse.keywords.KEYWORDS)
    [ret.append((kwd, _(u'Keyword')))
                for kwd in kwds]
    return ret


def find_matches(completions, fragment):
    """Find matches in the list of completions.

    The function returns a dictionary with the score determined by
    :meth:`get_score` as keys. The value is a list of entries in the form
    (indexes, (solution, description)) where *indexes* is again a 2-tuple
    defining the start and end index that's needed from *solution* to
    complete *fragment*, *solution* and *description* is the 2-tuple
    as returned by :meth:`build_completions`.
    """
    ret = {}
    for solution, comment in completions:
        score, indexes = get_score(solution, fragment)
        if score is None:
            continue
        tmp = ret.get(score, [])
        tmp.append((indexes, (solution, comment)))
        ret[score] = tmp
    return ret


def find_identifier(parsed, name=None):
    """Return the real name for an aliased identifier in a statement.

    :param parsed: Parsed SQL statement.
    :param name: Name to find.
    """

    def _collect_identifiers(tlist):
        for item in tlist:
            if isinstance(item, sqlparse.engine.grouping.Identifier):
                real_name = item.get_real_name()
                yield (item.get_alias() or real_name, real_name)
            elif item.is_group():
                for item in _collect_identifiers(item.tokens):
                    yield item

    # Cleanup aliases in temporary dict.
    # As we return the first match, we must take care that aliased names
    # are resolved.
    items = {}
    for alias, real_name in _collect_identifiers(parsed.tokens):
        if (alias in items
            and items[alias] == alias):
            items[alias] = real_name
        elif alias not in items:
            items[alias] = real_name
    if name is None:
        return items
    # Now return first match from dict.
    for alias, real_name in items.iteritems():
        if alias == name:
            return real_name


def get_score(solution, fragment):
    """Get score for solution."""
    solution = solution.lower()
    fragment = fragment.lower()
    try:
        start = solution.index(fragment)
    except ValueError:
        return None, None
    return start, [start+i for i in range(len(fragment))]


def get_completions_from_identifiers(editor):
    """Similar to get_completions but only with children of an identifier."""
    bounds = editor.textview.get_current_statement()
    meta = editor.connection.meta
    buffer_ = editor.textview.buffer
    completions = []
    if bounds:
        parsed = sqlparse.parse(buffer_.get_text(*bounds))[0]
        for alias, real_name in find_identifier(parsed).iteritems():
            for obj in meta.find(name=real_name):
                if obj.typeid in ('table', 'view'):
                    [completions.append((c.name, c.typestr))
                     for c in obj.columns.get_children()]
    return completions


def get_matches(editor):
    """Get possible completions."""
    buffer_ = editor.textview.buffer
    start, end = get_fragment_bounds(buffer_)
    fragment = buffer_.get_text(start, end)
    completions = None
    matches = None
    if len(fragment) == 0 and editor.connection and editor.connection.meta:
        completions = get_completions_from_identifiers(editor)
        matches = find_matches(completions, '')
    elif len(fragment) < 2:
        popup = editor.textview.get_data('cf::ac_window')
        if popup is not None:
            destroy_popup(editor.textview, popup)
        return
    if '.' in fragment and editor.connection and editor.connection.meta:
        parent_name, rest = fragment.split('.', 1)
        bounds = editor.textview.get_current_statement()
        if bounds:
            parsed = sqlparse.parse(buffer_.get_text(*bounds))[0]
            parent = find_identifier(parsed, parent_name)
            for obj in editor.connection.meta.find(name=parent):
                if obj is not None:
                    if obj.typeid == 'table':
                        completions = [(c.name, c.typestr)
                                       for c in obj.columns.get_children()]
                    elif obj.typeid == 'schema':
                        completions = [(c.name, c.typestr)
                                       for c in obj.tables.get_children()+
                                   obj.views.get_children()]
                if completions:
                    matches = find_matches(completions, rest)
    if completions is None:
        completions = build_completions(editor, fragment)
        matches = find_matches(completions, fragment)
    return matches


def get_fragment_bounds(buffer_, replace_mode=False, value=None):
    """Return start and end iter for fragment bounds.

    *end* is usually the current cursor position and *start* is much
    like backward_word_start() except that it respects some SQL specific
    stuff.
    """
    end = buffer_.get_iter_at_mark(buffer_.get_insert())
    start = end.copy()
    prev = start.copy()
    dot_start = None
    dot_count = 0
    while not prev.starts_line() and not prev.is_start():
        prev.backward_char()
        char = buffer_.get_text(prev, start)
        if char.isalnum() or char in '._"':
            if char == '.' and replace_mode:
                dot_count += 1
                if value is not None and dot_count > value.count("."):
                    return start, end
            start.backward_char()
        else:
            break
    return start, end


def _apply_selection(editor, value, add_blank=False):
    """Helper function that does the modifications in the text buffer."""
    if add_blank:
        value = '%s ' % value
    buffer_ = editor.textview.buffer
    start, end = get_fragment_bounds(buffer_, replace_mode=True, value=value)
    buffer_.begin_user_action()
    buffer_.delete(start, end)
    buffer_.insert(start, value)
    buffer_.end_user_action()

# ----
# Callbacks
# ----

def editor_key_pressed(textview, event, editor, window):
    """Handle editor's key-press-event when popup is visible."""
    if event.keyval == gtk.keysyms.Escape:
        destroy_popup(textview, window)
        textview.stop_emission('key-press-event')
        return True
    elif event.keyval == gtk.keysyms.Down:
        select_offset(window, 1)
        textview.stop_emission('key-press-event')
        return True
    elif event.keyval == gtk.keysyms.Up:
        select_offset(window, -1)
        textview.stop_emission('key-press-event')
        return True
    elif event.keyval == gtk.keysyms.Page_Down:
        select_offset(window, 5)
        textview.stop_emission('key-press-event')
        return True
    elif event.keyval == gtk.keysyms.Page_Up:
        select_offset(window, -5)
        textview.stop_emission('key-press-event')
        return True
    elif event.keyval == gtk.keysyms.Return:
        apply_selection(window, editor)
        textview.stop_emission('key-press-event')
        return True
    gobject.idle_add(editor_autocomplete, editor, window)
    return False


def on_editor_key_pressed_tab(sqlview, event, editor):
    """Tab key was pressed. Should only occur in advanced mode."""
    if event.keyval != gtk.keysyms.Tab:
        return False
    editor_autocomplete_advanced(editor)
    return True


def on_editor_created(instance, editor):
    """Connects to key-press-event to track TABs in advanced mode."""
    if instance.app.config.get('editor.tabcompletion'):
        if editor.get_data('cf::ac_tab') is not None:  # already connected
            return
        sig = editor.textview.connect('key-press-event',
                                      on_editor_key_pressed_tab, editor)
        editor.set_data('cf::ac_tab', sig)


def on_config_changed(config, key, enabled):
    """Tracks changes of editor.tabcompletion."""
    if key == 'editor.tabcompletion':
        for instance in config.app.get_instances():
            for editor in instance.get_editors():
                if enabled:
                    on_editor_created(instance, editor)
                elif not enabled and editor.get_data('cf::ac_tab'):
                    editor.textview.disconnect(editor.get_data('cf::ac_tab'))
                    editor.set_data('cf::ac_tab', None)


UI = """<menubar name="MenuBar">
  <menu name="Query" action="query-menu-action">
    <placeholder name="query-extensions">
      <menuitem name="query-autocomplete" action="query-autocomplete" />
    </placeholder>
  </menu>
</menubar>"""



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""UI helper functions."""

import os

import gtk

from cf import DATA_DIR


def get_logo_icon_list():
    """Returns a list of pixbufs for use with gtk.Window.set_icon_list()."""
    return tuple(get_logo_icon(size) for size in [24, 48, 96, 256])

def get_logo_icon(size):
    """Returns a pixbuf with the logo."""
    icon_dir = os.path.join(DATA_DIR, 'pixmaps/')
    fname = 'crunchyfrog_%dx%d.png' % (size, size)
    return gtk.gdk.pixbuf_new_from_file(os.path.join(icon_dir, fname))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from distutils.command.build import build
from distutils.command.clean import clean
from distutils.command.install import install as _install
from distutils.core import setup
from glob import glob
import os
import shutil
import sys

CMD_CLASS = {}

try:
    from sphinx.setup_command import BuildDoc
    CMD_CLASS['build_manual'] = BuildDoc
    CMD_CLASS['build_devguide'] = BuildDoc
except ImportError:
    pass

from utils.command.build_api import build_api
from utils.command.build_manpage import build_manpage
from utils.command.build_mo import build_mo
from utils.command.clean_docs import clean_docs
from utils.command.clean_mo import clean_mo


from cf import release


def find_packages(base):
    """Helper function to find subpackages in base.

    :param base: Directory holding the main package.

    Returns a list of subpackages suitable for setup() packages keyword.
    base is already included in the returned list.
    """
    ret = [base]
    for path in os.listdir(base):
        if path.startswith('.'):
            continue
        full_path = os.path.join(base, path)
        if os.path.isdir(full_path):
            ret += find_packages(full_path)
    return ret


class install(_install):
    """install command -- generates a local_config.py"""

    def run(self):
        if sys.platform == 'linux2':
            destdir = os.getenv('DESTDIR', '')
            if destdir and self.install_data.startswith(destdir):
                prefix = self.install_data[len(destdir):]
            else:
                prefix = self.install_data
            if prefix[0] != '/':
                prefix = '/%s' % prefix
        else:
            prefix = self.install_data
        f = open('cf/local_config.py', 'w')
        f.write('# -*- coding: utf-8 -*-\n\n')
        f.write('PREFIX = \'%s\'\n' % prefix)
        f.close()
        _install.run(self)

CMD_CLASS['install'] = install


class clean_with_subcommands(clean):

    def run(self):
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)
        clean.run(self)


CMD_CLASS['clean'] = clean_with_subcommands
CMD_CLASS['clean_docs'] = clean_docs
CMD_CLASS['build_api'] = build_api
CMD_CLASS['build_manpage'] = build_manpage
CMD_CLASS['build_mo'] = build_mo
CMD_CLASS['clean_mo'] = clean_mo


if 'build_manual' in CMD_CLASS:
    build.sub_commands.append(('build_manual', None))


DATA_FILES = []

# Glade
DATA_FILES += [('share/crunchyfrog/glade', glob('data/glade/*.glade'))]

# Plugins
# Expects that each directory in data/plugins is a plugin and that
# the plugin directories contain Python files only.
for dirname in os.listdir('data/plugins'):
    if dirname[0] in ('.', '_'):
        continue
    DATA_FILES += [(os.path.join('share/crunchyfrog/plugins', dirname),
                    glob(os.path.join('data/plugins', dirname, '*.py')))]

# Pixmaps
DATA_FILES += [('share/crunchyfrog/pixmaps', glob('data/pixmaps/*.png'))]

# Data
DATA_FILES += [('share/applications', ['data/crunchyfrog.desktop'])]
DATA_FILES += [('share/icons/hicolor/scalable/apps', ['data/crunchyfrog.svg'])]
DATA_FILES += [('share/pixmaps', ['data/crunchyfrog.svg'])]
DATA_FILES += [('share/icons/hicolor/24x24/apps', ['data/crunchyfrog.png'])]

# Manpage
DATA_FILES += [('share/man/man1', ['data/crunchyfrog.1'])]

# Documentation
DATA_FILES += [('share/doc/crunchyfrog', ['README', 'CHANGES', 'COPYING'])]
DATA_FILES += [('share/doc/crunchyfrog/manual',
                glob('docs/manual/build/html/*.*')),
               ('share/doc/crunchyfrog/manual/_images',
                glob('docs/manual/build/html/_images/*.*')),
               ('share/doc/crunchyfrog/manual/_sources',
                glob('docs/manual/build/html/_sources/*.*')),
               ('share/doc/crunchyfrog/manual/_static',
                glob('docs/manual/build/html/_static/*.*'))]

# Locales
for lang in os.listdir('po/'):
    path = os.path.join('po/', lang)
    if not os.path.isdir(path) or lang[0] in ['.', '_']:
        continue
    DATA_FILES += [('share/locale/%s/LC_MESSAGES' % lang,
                    ['po/%s/LC_MESSAGES/crunchyfrog.mo' % lang])]


setup(
    name=release.appname,
    version=release.version,
    description=release.description,
    author=release.author,
    author_email=release.author_email,
    long_description=release.long_description,
    license='GPL',
    url=release.url,
    download_url='http://crunchyfrog.googlecode.com/files/crunchyfrog-%s.tar.gz' % release.version,
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: X11 Applications :: GTK',
        'Environment :: X11 Applications :: Gnome',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Database :: Front-Ends',
    ],
    packages=find_packages('cf'),
    package_data={'cf': ['config/default.cfg']},
    scripts=['crunchyfrog'],
    data_files=DATA_FILES,
    cmdclass=CMD_CLASS,
)



# -*- coding: utf-8 -*-

"""Cleanup .mo files."""

from distutils.command.clean import clean
from distutils.core import Command
import os


def _get_languages():
    for lang in os.listdir('po'):
        if lang.endswith('.pot') or lang[0] in ['.', '_']:
                continue
        yield lang

class clean_mo(Command):

    description = 'Clean up generated .mo files.'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for lang in _get_languages():
            mofile = os.path.join('po/%s/LC_MESSAGES/crunchyfrog.mo' % lang)
            if os.path.isfile(mofile):
                print 'cleaning %s' % mofile
                os.remove(mofile)


clean.sub_commands.append(('clean_mo', None))



# -*- coding: utf-8 -*-

"""build_mo command to compile gettext catalogs."""

from distutils.command.build import build
from distutils.core import Command

import os
import sys

from utils import msgfmt

PO_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                       '../../po'))


class build_mo(Command):

    description = 'Compile message catalogs'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def _generate(self, lang):
        msgfmt.MESSAGES = {}  # reset cache (see issue96)
        infile = os.path.join(PO_PATH, lang, 'LC_MESSAGES', 'crunchyfrog.po')
        outfile = os.path.join(PO_PATH, lang, 'LC_MESSAGES', 'crunchyfrog.mo')
        msgfmt.make(infile, outfile)

    def run(self):
        for path in os.listdir(PO_PATH):
            if os.path.isdir(os.path.join(PO_PATH, path)):
                self._generate(path)


build.sub_commands.append(('build_mo', None))



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2009 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Native Shells"""

from gettext import gettext as _

import gtk

from cf.db import Datasource
from cf.plugins.core import GenericPlugin
from cf.plugins.mixins import InstanceMixin
from cf.ui import dialogs
from cf.ui.pane import PaneItem


class NativeShellPlugin(GenericPlugin, InstanceMixin):

    id = "crunchyfrog.plugin.nativeshell"
    name = _(u"Native Shell")
    description = _(u"Native Database shells")
    icon = "gnome-terminal"
    author = "Andi Albrecht"
    license = "GPL"
    homepage = "http://crunchyfrog.googlecode.com"
    version = "0.1"

    def __init__(self, app):
        self._instances = dict()
        self.app = app

    def init_instance(self, instance):
        if instance in self._instances:
            return
        sid = instance.browser.connect('object-menu-popup',
                                       self.on_object_menu_popup)
        self._instances[instance] = (sid, [])

    def shutdown(self):
        while self._instances:
            instance, args = self._instances.popitem()
            sid, views = args
            for view in views:
                view.destroy()
            instance.browser.disconnect(sid)

    def on_object_menu_popup(self, browser, popup, obj, path):
        if not isinstance(obj, Datasource):
            return
        item = gtk.MenuItem(_(u'Open Native Shell'))
        item.show()
        popup.append(item)
        cmd = obj.backend.get_native_shell_command(obj.url)
        if cmd is None:
            item.set_sensitive(False)
        else:
            item.set_sensitive(True)
            item.connect('activate', self.on_start_shell, obj,
                         browser.instance)

    def on_start_shell(self, menuitem, obj, instance):
        view = NativeShell(self.app, instance, obj)
        instance.queries.add_item(view)


try:
    import vte
except ImportError, err:
    NativeShellPlugin.INIT_ERROR = _(u'Please install python-vte.')
    class Dummy(object):
        Terminal = str
    vte = Dummy



class NativeShell(gtk.ScrolledWindow, PaneItem):

    name = _(u'Native Shell')
    icon = 'gnome-terminal'
    detachable = True
    connection = None  # For compatibility with Editor widget

    def __init__(self, app, instance, datasource):
        gtk.ScrolledWindow.__init__(self)
        PaneItem.__init__(self, app)
        self.app = app
        self.instance = instance
        self.datasource = datasource
        self.tab_label.set_datasource(datasource)
        self.tab_label.set_text(_(u'Shell') + (': %s' % datasource.public_url))
        self.term = vte.Terminal()
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.add(self.term)
        self.term.show()
        cmd, args = self.datasource.backend.get_native_shell_command(
            self.datasource.url)
        args.insert(0, cmd)
        self.term.connect('child-exited', self.on_child_exited, cmd, args)
        self.term.fork_command(cmd, args)

    def on_child_exited(self, term, cmd, args):
        exit_status = term.get_child_exit_status()
        if exit_status != 0:
            msg = _((u'The command "%(command)s" failed '
                     u'(Exit status: %(status)d).'))
            msg = msg % {'command': cmd, 'status': exit_status}
            gtk.gdk.threads_enter()
            dialogs.error(_(u'Failed'), msg)
            gtk.gdk.threads_leave()
        self.destroy()

    def get_focus_child(self):
        return self.term



# -*- coding: utf-8 -*-

"""Cleanup generated documentation files."""

import os
import shutil
from distutils.command.clean import clean
from distutils.core import Command


class clean_docs(Command):

    description = 'Clean up generated documentation files.'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for path in ['docs/devguide/build', 'docs/manual/build']:
            print 'Cleaning %s' % path
            for item in os.listdir(path):
                if item == '.DUMMY':
                    continue
                shutil.rmtree(os.path.join(path, item))


clean.sub_commands.append(('clean_docs', None))



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple message dialogs."""

from gettext import gettext as _

import gobject
import gtk


DIALOG_TITLES = {
    gtk.STOCK_DIALOG_ERROR: _(u'An error occurred'),
    gtk.STOCK_DIALOG_WARNING: _(u'Warning'),
    gtk.STOCK_DIALOG_INFO: _(u'Information'),
}


def _run_dialog(short, long=None, parent=None, buttons=gtk.BUTTONS_NONE,
                dlg_type=gtk.STOCK_DIALOG_INFO, title=None):
    if dlg_type == gtk.STOCK_DIALOG_ERROR:
        dlg_type2 = gtk.MESSAGE_ERROR
    elif dlg_type == gtk.STOCK_DIALOG_WARNING:
        dlg_type2 = gtk.MESSAGE_WARNING
    elif dlg_type == gtk.STOCK_DIALOG_QUESTION:
        dlg_type2 = gtk.MESSAGE_QUESTION
    else:
        dlg_type2 = gtk.MESSAGE_INFO
    if title is None:
        title = DIALOG_TITLES.get(dlg_type, 'CrunchyFrog')
    dlg = gtk.MessageDialog(parent,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            dlg_type2, buttons)
    short = gobject.markup_escape_text(short)
    dlg.set_markup('<b>%s</b>' % short)
    if long is not None:
        dlg.format_secondary_markup(gobject.markup_escape_text(long))
    img = gtk.image_new_from_stock(dlg_type, gtk.ICON_SIZE_DIALOG)
    dlg.set_image(img)
    dlg.set_title(title)
    dlg.set_icon_name('crunchyfrog')
    img.show()
    ret = dlg.run()
    dlg.destroy()
    return ret


def error(short, long=None, parent=None, buttons=gtk.BUTTONS_OK):
    return _run_dialog(short, long, parent, buttons, gtk.STOCK_DIALOG_ERROR)

def warning(short, long=None, parent=None, buttons=gtk.BUTTONS_OK):
    return _run_dialog(short, long, parent, buttons, gtk.STOCK_DIALOG_WARNING)

def yesno(short, long=None, parent=None, buttons=gtk.BUTTONS_YES_NO):
    return _run_dialog(short, long, parent, buttons, gtk.STOCK_DIALOG_QUESTION)

def info(short, long=None, parent=None, buttons=gtk.BUTTONS_OK):
    return _run_dialog(short, long, parent, buttons, gtk.STOCK_DIALOG_INFO)

def password(title, msg=None, parent=None):
    dlg = gtk.MessageDialog(parent,
                            gtk.DIALOG_MODAL|gtk.DIALOG_DESTROY_WITH_PARENT,
                            gtk.MESSAGE_QUESTION,
                            gtk.BUTTONS_OK_CANCEL)
    title = gobject.markup_escape_text(title)
    dlg.set_markup('<b>%s</b>' % title)
    dlg.set_title(_(u'Password required'))
    dlg.set_icon_name('crunchyfrog')
    if msg is not None:
        dlg.format_secondary_markup(msg)
    hbox = gtk.HBox()
    hbox.set_border_width(6)
    hbox.set_spacing(7)
    hbox.show()
    dlg.vbox.pack_start(hbox)

    label = gtk.Label(_('Password:'))
    label.show()
    hbox.pack_start(label, False, False)

    entry = gtk.Entry()
    entry.set_invisible_char(u'\u2022')
    entry.set_visibility(False)
    entry.show()
    hbox.pack_start(entry)

    if dlg.run() == gtk.RESPONSE_OK:
        ret = entry.get_text()
    else:
        ret = None
    dlg.destroy()
    return ret
